using System;
using System.Drawing;
using System.Windows.Forms;
using PSVRender;
using System.Drawing.Imaging;
using FluxShared;
using WeifenLuo.WinFormsUI.Docking;



namespace MapPlugin {
	public class LocationForm : MapForm {
		byte[] L12AnimatedSubtiles = new byte[0x2000];
		byte nScrollLeft = 0;
		byte nScrollRight = 0;
		byte nScrollTop = 0;
		byte nScrollBottom = 0;

		
		
		public LocationForm() : base(0){
			InitializeComponent();

			TileMap = new cLocMap(this);
			mnuMusicTransition.Enabled = false;
			mnuNLZ.Enabled = false;
		}



		protected override void Dispose(bool disposing){
			if(disposing){
				if(components != null){
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}

		
		
		#region Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			components = new System.ComponentModel.Container();
		}
		#endregion

		
	
		public override void WriteToMemory(uint nIndex){
			if(nIndex < 0xFFFF){
				SaveRecord Rec = G.SaveRec[(byte) RecType.LocProp][(int) nIndex];
				Rec.nData[1] = (byte) G.LocPropForm.L12TilesetSpin.Value;
				Rec.nData[2] = (byte) G.LocPropForm.L3TilesetSpin.Value;
				Rec.nData[3] = (byte) G.LocPropForm.PaletteSpin.Value;
				Rec.nData[4] = (byte) G.LocPropForm.MapSpin.Value;
				Rec.nData[10] = (byte) G.LocPropForm.LeftSpin.Value;
				Rec.nData[11] = (byte) G.LocPropForm.TopSpin.Value;
				Rec.nData[12] = (byte) G.LocPropForm.RightSpin.Value;
				Rec.nData[13] = (byte) G.LocPropForm.BottomSpin.Value;

				G.PostStatus("Writing tile properties");
				if(G.nRomType != 2) {
					CompressionDataVar RecVar = new CompressionDataVar();
					RecVar.nDecompressedSize = (uint) (3 * TileMap.nMaxHeight * TileMap.nMaxWidth);
					RecVar.nDecRoutine = CompressionRoutines.eChronoTriggerRunLength;
					RecVar.nSrcOff = 0;
					RecVar.nWorkOff = TileMap.MapTilePropOff;
					RecVar.SrcBuffer = TileMap.MapTileProp;
					RecVar.WorkingBuffer = TileMap.MapData;
					RecVar.CompressData();

					G.SaveRec[(byte) RecType.LocMap][Rec.nData[4]].nDataSize = (ushort) (TileMap.MapTilePropOff + RecVar.nCompressedSize);
				}
				else {
					uint nMapSize = (uint) (TileMap.nMaxHeight * TileMap.nMaxWidth);
					for(int i = 0; i < nMapSize; i++) {
						TileMap.MapData[i + TileMap.MapTilePropOff] = TileMap.MapTileProp[i * 3];
						TileMap.MapData[i + TileMap.MapTilePropOff + nMapSize] = TileMap.MapTileProp[(i * 3) + 1];
						TileMap.MapData[i + TileMap.MapTilePropOff + (nMapSize << 1)] = TileMap.MapTileProp[(i * 3) + 2];
					}

					G.SaveRec[(byte) RecType.LocMap][Rec.nData[4]].nDataSize = (ushort) (TileMap.MapTilePropOff + (nMapSize * 3));
				}

				Array.Copy(TileMap.MapData, 0, G.SaveRec[(byte) RecType.LocMap][Rec.nData[4]].nData, 0, G.SaveRec[(byte) RecType.LocMap][Rec.nData[4]].nDataSize);
				G.SaveRec[(byte) RecType.LocMap][Rec.nData[4]].bModified = true;
			}
			G.PostStatus("Done");
		}

	
	
		public override void ChangeTileset12(){
			if(G.LocPropForm.L12TilesetSpin.Value < 0x40) {
				uint i;

				G.PostStatus("Decompressing graphics");
				for(i = 0; i < 8; i++) {
					nTileset[i] = G.SaveRec[(byte) RecType.LocTileset][(int) G.LocPropForm.L12TilesetSpin.Value].nData[i];
					if(nTileset[i] < 0xE0) {
						if(i < 5) {
							Array.Copy(G.SaveRec[(int) RecType.LocTilesL12][nTileset[i]].nData, 0, L12SubtileData, i << 12, 0x1000);
						}
						else if(i == 5) {
							Array.Copy(G.SaveRec[(int) RecType.LocTilesL12][nTileset[i]].nData, 0, L12SubtileData, i << 12, 0xF80);
						}
						else {
							Array.Copy(G.SaveRec[(int) RecType.LocTilesL12][nTileset[i]].nData, 0, L12AnimatedSubtiles, (i - 6) << 12, 0x1000);
						}
					}
				}
			}
			else {
				Array.Clear(L12SubtileData, 0, L12SubtileData.Length);
			}

			G.PostStatus("Drawing subtiles");
			bmpL12Subtiles = new Bitmap(128, 384);
			ChunkVar12.WorkingImage = bmpL12Subtiles;
			ChunkVar12.nDecodingStart = 0;
			ChunkVar12.ProcessSpriteData();

			if(!bNoUpdate) {
				DrawL12Tiles();
			}
		}

		public override void DrawL12Tiles() {
			bmpL12Tiles = new Bitmap(256, 512);
			if(G.LocPropForm.L12TilesetSpin.Value < 0x40) {
				L12TileAsm = G.SaveRec[(byte) RecType.LocTileAsmL12][(byte) G.LocPropForm.L12TilesetSpin.Value].nData;

				GetSnesPaletteVar SnesPalVar = new GetSnesPaletteVar();
				SnesPalVar.nPaletteColors = 15;
				SnesPalVar.nPaletteShift = 16;
				SnesPalVar.nSnesStartIndex = 0;
				SnesPalVar.WorkingPalette = LPalette;
				SnesPalVar.nWorkStartIndex = 0;
				SnesPalVar.nPaletteSize = 16;
				SnesPalVar.nNumPalettes = 1;
				SnesPalVar.GetSnesPalette();

				SnesPalVar.nSnesPalette = G.SaveRec[(int) RecType.LocPal][(uint) G.LocPropForm.PaletteSpin.Value].nData;
				SnesPalVar.nNumPalettes = 7;
				SnesPalVar.nPaletteShift = 1;
				SnesPalVar.nWorkStartIndex = 0x10;
				SnesPalVar.nSnesStartIndex = 0;
				SnesPalVar.GetSnesPalette();

				ushort nSubtile;
				PutSubtileVar TileChunk = new PutSubtileVar();
				TileChunk.nBitPlanes = BitPlanes.eTetraPlane;
				TileChunk.WorkingBuffer = L12SubtileData;
				TileChunk.WorkingImage = bmpL12Tiles;
				TileChunk.WorkingPalette = LPalette;

				G.PostStatus("Drawing tiles");
				unsafe {
					BitmapData WIData = bmpL12Tiles.LockBits(new Rectangle(0, 0, bmpL12Tiles.Width, bmpL12Tiles.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
					TileChunk.nWorkingImage = (byte*) WIData.Scan0.ToPointer();
					for(int y = 0; y < 32; y++) {
						for(int x = 0; x < 16; x++) {
							for(int nChunk = 0; nChunk < 4; nChunk++) {
								nSubtile = G.GetShort(L12TileAsm, (((y << 4) | x) << 3) + (nChunk << 1));
								nTilePrior[nChunk | (x << 2) | (y << 6)] = (byte) ((nSubtile & 0x2000) >> 13);
								TileChunk.nDecodingStart = (ushort) (((nSubtile & 0x03FF) - 0x0100) << 5);
								if(TileChunk.nDecodingStart < 0x6000) {
									TileChunk.nFlip = (byte) (nSubtile >> 14);
									TileChunk.nPaletteShift = (byte) ((nSubtile & 0x1C00) >> 10);
									TileChunk.dx = (x << 4) | ((nChunk & 0x01) << 3);
									TileChunk.dy = (y << 4) | ((nChunk & 0x02) << 2);
									TileChunk.PutSubtile();
								}
							}
						}
					}
					bmpL12Tiles.UnlockBits(WIData);
				}
			}
			else {
				L12TileAsm = new byte[0x1000];
			}
			G.L12TilesForm.SetTiles(bmpL12Tiles);
		}

	
	
		protected override void OnFocusEnter(object sender, System.EventArgs e) {
			G.LocPropForm.ActiveMap = this;
			if(!bNoUpdate){
				bNoUpdate = true;
				G.LocPropForm.LocationCombo.SelectedIndex = nLocation;
				G.LocPropForm.EventsSpin.Value = nEvents;
				G.LocPropForm.MapSpin.Value = nMap;
				G.LocPropForm.LeftSpin.Value = nScrollLeft;
				G.LocPropForm.RightSpin.Value = nScrollRight;
				G.LocPropForm.TopSpin.Value = nScrollTop;
				G.LocPropForm.BottomSpin.Value = nScrollBottom;
				G.LocPropForm.PaletteSpin.Value = nLPalette;
				G.LocPropForm.L12TilesetSpin.Value = nTileset[0];
				G.LocPropForm.L3TilesetSpin.Value = nL3Tileset;
				bNoUpdate = false;
			}
			G.LocPropForm.LocPropPanel.Enabled = true;
			G.L12TilesForm.SetTiles(bmpL12Tiles);
			G.L3TilesForm.SetTiles(bmpL3Tiles);
		}
	
	
	
		protected override void OnFocusLeave(object sender, System.EventArgs e){
			nLocation = (ushort) G.LocPropForm.LocationCombo.SelectedIndex;
			nEvents = (ushort) G.LocPropForm.EventsSpin.Value;
			nMap = (byte) G.LocPropForm.MapSpin.Value;
			nScrollLeft = (byte) G.LocPropForm.LeftSpin.Value;
			nScrollRight = (byte) G.LocPropForm.RightSpin.Value;
			nScrollTop = (byte) G.LocPropForm.TopSpin.Value;
			nScrollBottom = (byte) G.LocPropForm.BottomSpin.Value;
			nLPalette = (byte) G.LocPropForm.PaletteSpin.Value;
			nTileset[0] = (byte) G.LocPropForm.L12TilesetSpin.Value;
			nL3Tileset = (byte) G.LocPropForm.L3TilesetSpin.Value;
			G.LocPropForm.LocPropPanel.Enabled = false;
		}
	
	
	
		public override void ChangeTileset3(){
			if(G.LocPropForm.L3TilesetSpin.Value < 0x17) {
				G.PostStatus("Decompressing graphics");
				Array.Copy(G.SaveRec[(int) RecType.LocTilesL12][(int) G.LocPropForm.L3TilesetSpin.Value].nData, 0, L3SubtileData, 0, 0x1000);
			}
			else {
				Array.Clear(L3SubtileData, 0, L3SubtileData.Length);
			}

			bmpL3Subtiles = new Bitmap(128, 128);
			PutSubtileVar L3ChunkVar = new PutSubtileVar();
			L3ChunkVar.nBitPlanes = BitPlanes.eDiPlane;
			L3ChunkVar.WorkingBuffer = L3SubtileData;
			L3ChunkVar.WorkingImage = bmpL3Subtiles;
			L3ChunkVar.WorkingPalette = new Color[16];
			PSV.GetGreyPalette(ref L3ChunkVar.WorkingPalette, 16);

			G.PostStatus("Drawing subtiles");
			L3ChunkVar.nDecodingStart = 0;
			L3ChunkVar.ProcessSpriteData();

			if(!bNoUpdate) {
				DrawL3Tiles();
			}
		}

		public override void DrawL3Tiles() {
			bmpL3Tiles = new Bitmap(256, 256);
			if(G.LocPropForm.L3TilesetSpin.Value < 0x17) {
				L3TileAsm = G.SaveRec[(byte) RecType.LocTileAsmL3][(byte) G.LocPropForm.L3TilesetSpin.Value].nData;

				GetSnesPaletteVar SnesPalVar = new GetSnesPaletteVar();
				SnesPalVar.nPaletteColors = 15;
				SnesPalVar.nPaletteShift = 16;
				SnesPalVar.nSnesStartIndex = 0;
				SnesPalVar.WorkingPalette = LPalette;
				SnesPalVar.nWorkStartIndex = 0;
				SnesPalVar.nPaletteSize = 16;
				SnesPalVar.nNumPalettes = 1;
				SnesPalVar.GetSnesPalette();

				SnesPalVar.nSnesPalette = G.SaveRec[(int) RecType.LocPal][(uint) G.LocPropForm.PaletteSpin.Value].nData;
				SnesPalVar.nNumPalettes = 7;
				SnesPalVar.nPaletteShift = 1;
				SnesPalVar.nWorkStartIndex = 0x10;
				SnesPalVar.nSnesStartIndex = 0;
				SnesPalVar.GetSnesPalette();

				ushort nSubtile;
				PutSubtileVar TileChunk = new PutSubtileVar();
				TileChunk.nBitPlanes = BitPlanes.eDiPlane;
				TileChunk.WorkingBuffer = L3SubtileData;
				TileChunk.WorkingImage = bmpL3Tiles;
				TileChunk.WorkingPalette = LPalette;

				G.PostStatus("Drawing tiles");
				unsafe {
					BitmapData WIData = bmpL3Tiles.LockBits(new Rectangle(0, 0, bmpL3Tiles.Width, bmpL3Tiles.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
					TileChunk.nWorkingImage = (byte*) WIData.Scan0.ToPointer();
					for(int y = 0; y < 16; y++) {
						for(int x = 0; x < 16; x++) {
							for(int nChunk = 0; nChunk < 4; nChunk++) {
								nSubtile = G.GetShort(L3TileAsm, (((y << 4) | x) << 3) + (nChunk << 1));
								nTilePrior[nChunk | (x << 2) | (y << 6) | 0x800] = (byte) ((nSubtile & 0x2000) >> 13);
								TileChunk.nDecodingStart = (ushort) ((nSubtile & 0x03FF) << 4);
								if(TileChunk.nDecodingStart < 0x1000) {
									TileChunk.nFlip = (byte) (nSubtile >> 14);
									TileChunk.nPaletteShift = (byte) ((nSubtile & 0x1C00) >> 10);
									TileChunk.dx = (x << 4) | ((nChunk & 0x01) << 3);
									TileChunk.dy = (y << 4) | ((nChunk & 0x02) << 2);
									TileChunk.PutSubtile();
								}
							}
						}
					}
					bmpL3Tiles.UnlockBits(WIData);
				}
			}
			else {
				L3TileAsm = null;
			}
			G.L3TilesForm.SetTiles(bmpL3Tiles);
		}

		
		
		protected override void OnClose(object sender, FormClosingEventArgs e) {
			G.LocPropForm.ActiveMap = null;
		}
	}
}

