using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;
using System.Text;
using PSVRender;
using System.IO;
using GGRLib;
using WeifenLuo.WinFormsUI.Docking;
using FluxShared;



namespace MapPlugin{
	public abstract class MapForm : DockContent{
		public byte[] L12SubtileData;
		public byte[] L3SubtileData = new byte[0x1000];
		public byte[] L12TileAsm;
		public Color[] LPalette = new Color[128];
		public byte[] L3TileAsm;
		public byte[] nTilePrior;
		protected PutSubtileVar ChunkVar12 = new PutSubtileVar();
		protected Bitmap bmpL12Subtiles;
		protected Bitmap bmpL3Subtiles;
		protected Bitmap bmpL3Tiles;
		protected Bitmap bmpL12Tiles;
		public bool bNoUpdate = false;
		protected bool bMouseDown = false;
		public ushort nLocation = 0;
		protected ushort nEvents = 0;
		protected ushort nMap = 0;
		protected byte nLPalette = 0;
		protected byte[] nTileset = new byte[8];
		protected byte nL3Tileset = 0;
		protected long nResizeTime = 0;
		protected ushort nPrevMouseTileX = 0;
		protected ushort nPrevMouseTileY = 0;
		public MenuItem[] mnuLayer;
		byte nType;
		FlatCombo LocCombo;
		public cTileMap TileMap;
		protected System.ComponentModel.IContainer components;
		protected System.Windows.Forms.MenuItem mnuLocation;
		private System.Windows.Forms.MenuItem menuItem9;
		private System.Windows.Forms.MenuItem menuItem4;
		protected System.Windows.Forms.MenuItem mnuMap;
		private System.Windows.Forms.MenuItem menuItem7;
		private System.Windows.Forms.MainMenu LocMenu;
		public System.Windows.Forms.MenuItem mnuScrollMask;
		public System.Windows.Forms.MenuItem mnuExits;
		public System.Windows.Forms.PictureBox imgMap;
		public System.Windows.Forms.MenuItem mnuSolidity;
		public System.Windows.Forms.MenuItem mnuZPlane;
		public System.Windows.Forms.MenuItem mnuWind;
		public System.Windows.Forms.MenuItem mnuBattle;
		public System.Windows.Forms.MenuItem mnuNPC;
		public System.Windows.Forms.MenuItem mnuDoorTrigger;
		public System.Windows.Forms.MenuItem mnuUnk120;
		public MenuItem mnuTreasure;
		private MenuItem mnuExp;
		public MenuItem[] mnuExplicit = new MenuItem[7];
		private SaveFileDialog SaveFileDlg;
		public MenuItem mnuMusicTransition;
		bool[][] bDisplayState = new bool[(byte) DisplayMode.Count][];
		public MenuItem mnuNLZ;
		DisplayMode Display = DisplayMode.Normal;



		
		public MapForm(byte nTypeIn){
			nType = nTypeIn;

			InitializeComponent();
			ExtraInit();

			ChunkVar12.nBitPlanes = BitPlanes.eTetraPlane;
			ChunkVar12.WorkingBuffer = L12SubtileData;
			ChunkVar12.WorkingPalette = new Color[16];
			PSV.GetGreyPalette(ref ChunkVar12.WorkingPalette, 16);
		}



		void ExtraInit(){
			mnuLayer = new MenuItem[3];
			mnuLayer[0] = new MenuItem("Layer 1", new EventHandler(OnCheckLayer), Shortcut.Ctrl1);
			mnuLayer[1] = new MenuItem("Layer 2", new EventHandler(OnCheckLayer), Shortcut.Ctrl2);
			mnuLayer[2] = new MenuItem("Layer 3", new EventHandler(OnCheckLayer), Shortcut.Ctrl3);
			byte i;
			for(i = 0; i < 3; i++) {
				if(i < 2){
					mnuLayer[i].Checked = true;
				}
				mnuLayer[i].Index = i;
				mnuMap.MenuItems.Add(i, mnuLayer[i]);
			}

			mnuExplicit[0] = new MenuItem("Normal", new EventHandler(OnExplicit), Shortcut.CtrlShift0);
			mnuExplicit[1] = new MenuItem("Layer 1", new EventHandler(OnExplicit), Shortcut.CtrlShift1);
			mnuExplicit[2] = new MenuItem("Layer 2", new EventHandler(OnExplicit), Shortcut.CtrlShift2);
			mnuExplicit[3] = new MenuItem("Layer 3", new EventHandler(OnExplicit), Shortcut.CtrlShift3);
			mnuExplicit[4] = new MenuItem("Properties", new EventHandler(OnExplicit), Shortcut.CtrlShift4);
			mnuExplicit[5] = new MenuItem("Most Recent Copy", new EventHandler(OnExplicit), Shortcut.CtrlShift5);
			mnuExplicit[6] = new MenuItem("Swap L1 and L2", new EventHandler(OnExplicit), Shortcut.CtrlShift6);
			for (i = 0; i < 7; i++) {
				mnuExplicit[i].Index = i;
				mnuExp.MenuItems.Add(i, mnuExplicit[i]);
			}
			mnuExplicit[0].Checked = true;

			if(nType == 0){
				L12SubtileData = new byte[0x6000];
				nTilePrior = new byte[0xC00];
				LocCombo = G.LocPropForm.LocationCombo;
			}
			//else{
			//    L12SubtileData = new byte[0x7000];
			//    nTilePrior = new byte[0x800];
			//    LocCombo = MForm.OWPropForm.OverworldCombo;
			//}

			for(i = 0; i < (byte) DisplayMode.Count; i++) {
				bDisplayState[i] = new bool[0x0A];
			}
			bDisplayState[(byte) DisplayMode.Exits][0] = true;
			bDisplayState[(byte) DisplayMode.MTD][9] = true;
			bDisplayState[(byte) DisplayMode.Props][2] = true;
			bDisplayState[(byte) DisplayMode.Treasure][1] = true;
		}



		protected override void Dispose(bool disposing) {
			if(disposing){
				if (components != null){
					components.Dispose();
				}
			}
			base.Dispose(disposing);
		}

		
		
		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.LocMenu = new System.Windows.Forms.MainMenu(this.components);
			this.mnuLocation = new System.Windows.Forms.MenuItem();
			this.menuItem9 = new System.Windows.Forms.MenuItem();
			this.mnuMap = new System.Windows.Forms.MenuItem();
			this.mnuScrollMask = new System.Windows.Forms.MenuItem();
			this.mnuExits = new System.Windows.Forms.MenuItem();
			this.mnuTreasure = new System.Windows.Forms.MenuItem();
			this.mnuSolidity = new System.Windows.Forms.MenuItem();
			this.mnuWind = new System.Windows.Forms.MenuItem();
			this.mnuZPlane = new System.Windows.Forms.MenuItem();
			this.mnuBattle = new System.Windows.Forms.MenuItem();
			this.mnuNPC = new System.Windows.Forms.MenuItem();
			this.mnuDoorTrigger = new System.Windows.Forms.MenuItem();
			this.mnuUnk120 = new System.Windows.Forms.MenuItem();
			this.mnuMusicTransition = new System.Windows.Forms.MenuItem();
			this.mnuNLZ = new System.Windows.Forms.MenuItem();
			this.menuItem7 = new System.Windows.Forms.MenuItem();
			this.mnuExp = new System.Windows.Forms.MenuItem();
			this.menuItem4 = new System.Windows.Forms.MenuItem();
			this.imgMap = new System.Windows.Forms.PictureBox();
			this.SaveFileDlg = new System.Windows.Forms.SaveFileDialog();
			((System.ComponentModel.ISupportInitialize) (this.imgMap)).BeginInit();
			this.SuspendLayout();
			// 
			// LocMenu
			// 
			this.LocMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuLocation,
            this.mnuMap});
			// 
			// mnuLocation
			// 
			this.mnuLocation.Index = 0;
			this.mnuLocation.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.menuItem9});
			this.mnuLocation.Text = "Location";
			// 
			// menuItem9
			// 
			this.menuItem9.Index = 0;
			this.menuItem9.Shortcut = System.Windows.Forms.Shortcut.CtrlS;
			this.menuItem9.Text = "Write To Memory";
			this.menuItem9.Click += new System.EventHandler(this.OnWriteToMemory);
			// 
			// mnuMap
			// 
			this.mnuMap.Index = 1;
			this.mnuMap.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuScrollMask,
            this.mnuExits,
            this.mnuTreasure,
            this.mnuSolidity,
            this.mnuWind,
            this.mnuZPlane,
            this.mnuBattle,
            this.mnuNPC,
            this.mnuDoorTrigger,
            this.mnuUnk120,
            this.mnuMusicTransition,
            this.mnuNLZ,
            this.menuItem7,
            this.mnuExp,
            this.menuItem4});
			this.mnuMap.Text = "Map";
			// 
			// mnuScrollMask
			// 
			this.mnuScrollMask.Checked = true;
			this.mnuScrollMask.Index = 0;
			this.mnuScrollMask.Shortcut = System.Windows.Forms.Shortcut.Ctrl4;
			this.mnuScrollMask.Text = "Scroll Mask";
			this.mnuScrollMask.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuExits
			// 
			this.mnuExits.Index = 1;
			this.mnuExits.Shortcut = System.Windows.Forms.Shortcut.Ctrl5;
			this.mnuExits.Text = "Exits";
			this.mnuExits.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuTreasure
			// 
			this.mnuTreasure.Index = 2;
			this.mnuTreasure.Shortcut = System.Windows.Forms.Shortcut.CtrlT;
			this.mnuTreasure.Text = "Treasure";
			this.mnuTreasure.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuSolidity
			// 
			this.mnuSolidity.Index = 3;
			this.mnuSolidity.Shortcut = System.Windows.Forms.Shortcut.Ctrl6;
			this.mnuSolidity.Text = "Solidity";
			this.mnuSolidity.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuWind
			// 
			this.mnuWind.Index = 4;
			this.mnuWind.Shortcut = System.Windows.Forms.Shortcut.Ctrl7;
			this.mnuWind.Text = "Wind";
			this.mnuWind.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuZPlane
			// 
			this.mnuZPlane.Index = 5;
			this.mnuZPlane.Shortcut = System.Windows.Forms.Shortcut.Ctrl8;
			this.mnuZPlane.Text = "Z-Plane";
			this.mnuZPlane.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuBattle
			// 
			this.mnuBattle.Index = 6;
			this.mnuBattle.Shortcut = System.Windows.Forms.Shortcut.Ctrl9;
			this.mnuBattle.Text = "Battle Solidity";
			this.mnuBattle.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuNPC
			// 
			this.mnuNPC.Index = 7;
			this.mnuNPC.Text = "NPC Solidity";
			this.mnuNPC.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuDoorTrigger
			// 
			this.mnuDoorTrigger.Index = 8;
			this.mnuDoorTrigger.Text = "Door Trigger";
			this.mnuDoorTrigger.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuUnk120
			// 
			this.mnuUnk120.Index = 9;
			this.mnuUnk120.Text = "Unknown 1.20";
			this.mnuUnk120.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuMusicTransition
			// 
			this.mnuMusicTransition.Index = 10;
			this.mnuMusicTransition.Text = "Music Transition";
			this.mnuMusicTransition.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// mnuNLZ
			// 
			this.mnuNLZ.Index = 11;
			this.mnuNLZ.Text = "No Landing Zone";
			this.mnuNLZ.Click += new System.EventHandler(this.OnCheckMenu);
			// 
			// menuItem7
			// 
			this.menuItem7.Index = 12;
			this.menuItem7.Text = "-";
			// 
			// mnuExp
			// 
			this.mnuExp.Index = 13;
			this.mnuExp.Text = "Explicit Paste";
			// 
			// menuItem4
			// 
			this.menuItem4.Index = 14;
			this.menuItem4.Shortcut = System.Windows.Forms.Shortcut.CtrlP;
			this.menuItem4.Text = "Save Map To PNG";
			this.menuItem4.Click += new System.EventHandler(this.OnSaveMapToPNG);
			// 
			// imgMap
			// 
			this.imgMap.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
			this.imgMap.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte) (0)));
			this.imgMap.Location = new System.Drawing.Point(1, 3);
			this.imgMap.Name = "imgMap";
			this.imgMap.Size = new System.Drawing.Size(404, 245);
			this.imgMap.SizeMode = System.Windows.Forms.PictureBoxSizeMode.StretchImage;
			this.imgMap.TabIndex = 1;
			this.imgMap.TabStop = false;
			this.imgMap.MouseMove += new System.Windows.Forms.MouseEventHandler(this.OnMapMouseMove);
			this.imgMap.MouseDown += new System.Windows.Forms.MouseEventHandler(this.OnMapMouseDown);
			this.imgMap.MouseUp += new System.Windows.Forms.MouseEventHandler(this.OnMapMouseUp);
			// 
			// SaveFileDlg
			// 
			this.SaveFileDlg.AddExtension = false;
			this.SaveFileDlg.DefaultExt = "txt";
			this.SaveFileDlg.FileName = "Chests.txt";
			this.SaveFileDlg.Filter = "Text Files (*.txt)|*.txt";
			this.SaveFileDlg.OverwritePrompt = false;
			// 
			// MapForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 14);
			this.AutoScroll = true;
			this.ClientSize = new System.Drawing.Size(406, 249);
			this.Controls.Add(this.imgMap);
			this.DockAreas = WeifenLuo.WinFormsUI.Docking.DockAreas.Document;
			this.Font = new System.Drawing.Font("Tahoma", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte) (0)));
			this.Menu = this.LocMenu;
			this.Name = "MapForm";
			this.ShowHint = WeifenLuo.WinFormsUI.Docking.DockState.Document;
			this.ShowInTaskbar = false;
			this.TabText = "Location";
			this.Text = "Location";
			this.Deactivate += new System.EventHandler(this.OnFocusLeave);
			this.Activated += new System.EventHandler(this.OnFocusEnter);
			this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.OnClose);
			this.Resize += new System.EventHandler(this.OnResize);
			((System.ComponentModel.ISupportInitialize) (this.imgMap)).EndInit();
			this.ResumeLayout(false);

		}
		#endregion

		
		
		protected void OnWriteToMemory(object sender, System.EventArgs e){
			WriteToMemory((uint) LocCombo.SelectedIndex);
		}


		
		public abstract void WriteToMemory(uint nIndex);


		
		public abstract void ChangeTileset12();



		public abstract void DrawL12Tiles();



		public abstract void ChangeTileset3();



		public abstract void DrawL3Tiles();

		
		
		public void ChangeMap() {
			TileMap.ChangeMap();
			if(!bNoUpdate) {
				TileMap.RedrawMap();
			}
		}



		public void ChangePalette(){
			if(!bNoUpdate) {
				DrawL3Tiles();
				DrawL12Tiles();
				TileMap.RedrawMap();
			}
		}



		private void OnSaveMapToPNG(object sender, System.EventArgs e) {
			string sFilename = "ct";
			if(nType == 0){
				sFilename += "loc";
			}
			else{
				sFilename += "ow";
			}
			sFilename += G.HexStr(LocCombo.SelectedIndex, 3) + ".png";
			imgMap.Image.Save(sFilename, System.Drawing.Imaging.ImageFormat.Png);
		}



		public void OnCheckLayer(object sender, System.EventArgs e){
			Cursor.Current = Cursors.WaitCursor;
			MenuItem mnuSender = (MenuItem) sender;
			mnuSender.Checked = !mnuSender.Checked;
			TileMap.RedrawMap();
			Cursor.Current = Cursors.Default;
		}



		public void OnExplicit(object sender, System.EventArgs e) {
			MenuItem mnuSender = (MenuItem) sender;
			if(mnuExplicit[0] == mnuSender) {
				mnuExplicit[0].Checked = true;
				mnuExplicit[1].Checked = false;
				mnuExplicit[2].Checked = false;
				mnuExplicit[3].Checked = false;
				mnuExplicit[4].Checked = false;
			}
			else{
				mnuSender.Checked = !mnuSender.Checked;
				if(mnuExplicit[1].Checked || mnuExplicit[2].Checked || mnuExplicit[3].Checked || mnuExplicit[4].Checked) {
					mnuExplicit[0].Checked = false;
				}
				else {
					mnuExplicit[0].Checked = true;
				}
			}
		}



		public void OnMapMouseUp(object sender, System.Windows.Forms.MouseEventArgs e){
			bMouseDown = false;
			long nOneSecondAgo = DateTime.Now.Ticks - 10000000;
			if(nResizeTime > nOneSecondAgo || G.nOpenTime > nOneSecondAgo) {
				return;
			}

			MouseEventArgs mea = new MouseEventArgs(e.Button, e.Clicks, e.X / G.nZoomFactor, e.Y / G.nZoomFactor, e.Delta);
			if(e.Button == MouseButtons.Left){
				TileMap.ChangeMapTile(mea);
			}
			else{
				TileMap.bCopyFinished = true;
				TileMap.SelectMapTile(mea);
				if(mnuLayer[2].Checked){
					G.L3TilesForm.OnMouseUp(null, mea);
				}
				else{
					G.L12TilesForm.OnMouseUp(null, mea);
				}
			}
		}
		

		
		public void OnMapMouseMove(object sender, System.Windows.Forms.MouseEventArgs e){
			MouseEventArgs mea = new MouseEventArgs(e.Button, e.Clicks, e.X / G.nZoomFactor, e.Y / G.nZoomFactor, e.Delta);
			
			G.CoordStatus.Text = "(" + mea.X.ToString("X") + ", " + mea.Y.ToString("X") + ")";
			
			ushort nMouseTileX = (ushort) (mea.X & 0xFFF0);
			ushort nMouseTileY = (ushort) (mea.Y & 0xFFF0);
			
			if(nMouseTileX == nPrevMouseTileX && nMouseTileY == nPrevMouseTileY){
				return;
			}
			
			TileMap.RestoreMap();
			Graphics graphMap = Graphics.FromImage(imgMap.Image);
			graphMap.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceOver;

			// Select where to get tile tracker from
			ushort[][][] SrcArea = TileMap.AreaSelected;
			ushort nSrcH = TileMap.nCopyAreaSelectH;
			ushort nSrcW = TileMap.nCopyAreaSelectW;

			if(nType == 0 && mnuExplicit[5].Checked) {
				SrcArea = G.RecentMap.AreaSelected;
				nSrcH = G.RecentMap.nCopyAreaSelectH;
				nSrcW = G.RecentMap.nCopyAreaSelectW;
			}


			if(bMouseDown){
				TileMap.bAreaSelectStarted = true;
				if(mea.Button == MouseButtons.Right){
					ushort nNewCopyAreaSelectW = 0;
					ushort nNewCopyAreaSelectH = 0;
					ushort nNewCopyAreaSelectX = 0;
					ushort nNewCopyAreaSelectY = 0;
					if(nMouseTileX < TileMap.nCopyAreaSelectX){
						nNewCopyAreaSelectW = (ushort) (TileMap.nCopyAreaSelectX - nMouseTileX);
						nNewCopyAreaSelectX = nMouseTileX;
					}
					else{
						nNewCopyAreaSelectW = (ushort) (nMouseTileX - TileMap.nCopyAreaSelectX);
						nNewCopyAreaSelectX = TileMap.nCopyAreaSelectX;
					}
					if(nMouseTileY < TileMap.nCopyAreaSelectY){
						nNewCopyAreaSelectH = (ushort) (TileMap.nCopyAreaSelectY - nMouseTileY);
						nNewCopyAreaSelectY = nMouseTileY;
					}
					else{
						nNewCopyAreaSelectH = (ushort) (nMouseTileY - TileMap.nCopyAreaSelectY);
						nNewCopyAreaSelectY = TileMap.nCopyAreaSelectY;
					}
					if(nNewCopyAreaSelectW != TileMap.nCopyAreaSelectW || nNewCopyAreaSelectH != TileMap.nCopyAreaSelectH){
						TileMap.nCopyAreaSelectW = nNewCopyAreaSelectW;
						TileMap.nCopyAreaSelectH = nNewCopyAreaSelectH;
						graphMap.FillRectangle(new SolidBrush(Color.FromArgb(0x20, 255, 255, 255)), nNewCopyAreaSelectX, nNewCopyAreaSelectY, nNewCopyAreaSelectW, nNewCopyAreaSelectH);
					}
				}
				else{
					ushort nNewPasteAreaSelectW = 0;
					ushort nNewPasteAreaSelectH = 0;
					ushort nNewPasteAreaSelectX = 0;
					ushort nNewPasteAreaSelectY = 0;
					if(nMouseTileX < TileMap.nPasteAreaSelectX){
						nNewPasteAreaSelectW = (ushort) (TileMap.nPasteAreaSelectX - nMouseTileX);
						nNewPasteAreaSelectX = nMouseTileX;
					}
					else{
						nNewPasteAreaSelectW = (ushort) (nMouseTileX - TileMap.nPasteAreaSelectX);
						nNewPasteAreaSelectX = TileMap.nPasteAreaSelectX;
					}
					if(nMouseTileY < TileMap.nPasteAreaSelectY){
						nNewPasteAreaSelectH = (ushort) (TileMap.nPasteAreaSelectY - nMouseTileY);
						nNewPasteAreaSelectY = nMouseTileY;
					}
					else{
						nNewPasteAreaSelectH = (ushort) (nMouseTileY - TileMap.nPasteAreaSelectY);
						nNewPasteAreaSelectY = TileMap.nPasteAreaSelectY;
					}
					if(nNewPasteAreaSelectW != TileMap.nPasteAreaSelectW || nNewPasteAreaSelectH != TileMap.nPasteAreaSelectH){
						TileMap.nPasteAreaSelectW = nNewPasteAreaSelectW;
						TileMap.nPasteAreaSelectH = nNewPasteAreaSelectH;
						graphMap.FillRectangle(new SolidBrush(Color.FromArgb(0x20, 255, 255, 255)), nNewPasteAreaSelectX, nNewPasteAreaSelectY, nNewPasteAreaSelectW, nNewPasteAreaSelectH);
					}
				}
			}
			else if(SrcArea != null && !(nSrcH == 0 || nSrcH > 64 || nSrcW == 0 || nSrcW > 64)){
				graphMap.FillRectangle(new SolidBrush(Color.FromArgb(0x20, 255, 255, 255)), nMouseTileX, nMouseTileY, (nSrcW << 4), (nSrcH << 4));
			}
			else{
				graphMap.FillRectangle(new SolidBrush(Color.FromArgb(0x20, 255, 255, 255)), nMouseTileX, nMouseTileY, 16, 16);
			}

			imgMap.Invalidate();
			nPrevMouseTileX = nMouseTileX;
			nPrevMouseTileY = nMouseTileY;
		}



		private void OnMapMouseDown(object sender, System.Windows.Forms.MouseEventArgs e) {
			MouseEventArgs mea = new MouseEventArgs(e.Button, e.Clicks, e.X / G.nZoomFactor, e.Y / G.nZoomFactor, e.Delta);
			bMouseDown = true;
			if(e.Button == MouseButtons.Right){
				TileMap.nCopyAreaSelectX = (ushort) (mea.X & 0xFFF0);
				TileMap.nCopyAreaSelectY = (ushort) (mea.Y & 0xFFF0);
				TileMap.nCopyAreaSelectW = 0x10;
				TileMap.nCopyAreaSelectH = 0x10;
				TileMap.bCopyFinished = false;
			}
			else{
				TileMap.nPasteAreaSelectX = (ushort) (mea.X & 0xFFF0);
				TileMap.nPasteAreaSelectY = (ushort) (mea.Y & 0xFFF0);
				TileMap.nPasteAreaSelectW = 0x10;
				TileMap.nPasteAreaSelectH = 0x10;
			}
		}



		protected abstract void OnFocusEnter(object sender, System.EventArgs e);


		
		protected abstract void OnFocusLeave(object sender, System.EventArgs e);



		protected void OnCheckMenu(object sender, System.EventArgs e){
			Cursor.Current = Cursors.WaitCursor;
			MenuItem mnuSender = (MenuItem) sender;
			mnuSender.Checked = !mnuSender.Checked;
			
			// Set check in display mode
			for(byte i = 0; i < 0x0A; i++) {
				if(mnuMap.MenuItems[4 + i] == mnuSender) {
					bDisplayState[(byte) Display][i] = mnuSender.Checked;
					break;
				}
			}

			TileMap.RestoreMap();
			Cursor.Current = Cursors.Default;
		}



		public void OnUpdateZoom(){
			imgMap.Height = (TileMap.bmpMap.Height + 4) * G.nZoomFactor;
			imgMap.Width = (TileMap.bmpMap.Width + 4) * G.nZoomFactor;
			imgMap.Refresh();
		}



		protected void OnResize(object sender, System.EventArgs e) {
			nResizeTime = DateTime.Now.Ticks;
		}

	
	
		public void ChangeMask(){
			if(!bNoUpdate){
				Cursor.Current = Cursors.WaitCursor;
				TileMap.RestoreMap();
				//TileMap.RestoreMap();
				Cursor.Current = Cursors.Default;
			}
		}


		
		protected abstract void OnClose(object sender, FormClosingEventArgs e);



		public void SetDisplayMode(DisplayMode nDisplay) {
			Cursor.Current = Cursors.WaitCursor;

			Display = nDisplay;
			for(byte i = 0; i < 0x0A; i++) {
				mnuMap.MenuItems[4 + i].Checked = bDisplayState[(byte) nDisplay][i];
			}
			
			TileMap.RestoreMap();
			Cursor.Current = Cursors.Default;
		}
	}



	public enum DisplayMode : byte {
		Normal,
		Exits,
		MTD,
		Props,
		Treasure,
		Count
	}
}
