using System;
using System.Collections.Generic;
using System.Text;
using FluxShared;
using System.Windows.Forms;
using WeifenLuo.WinFormsUI.Docking;

namespace MapPlugin {
	public class MapPlugin : IFluxPlugin{
		MenuItem mnuMap;
		MenuItem mnuSub;

		public List<SaveRecord[]> RecList {
			get {
				return G.SaveRec;
			}
			set {
				G.SaveRec = value;
			}
		}

		public MenuItem PlugMenu{
			get{
				return mnuMap;
			}
			set{
				mnuMap = value;
			}
		}

		public string sPlugName {
			get {
				return "Map Plug";
			}
		}

		public ushort nFluxSchema {
			get {
				return 0x0001;
			}
		}

		public ushort nFluxVersion {
			get {
				return 0x0100;
			}
		}

		public ushort nFluxMinSchema {
			get {
				return 0x0001;
			}
		}

		public ushort nFluxMinVersion {
			get {
				return 0x0100;
			}
		}

		public Dictionary<string, int> RecDict {
			get {
				return G.RecTypeDict;
			}
			set {
				G.RecTypeDict = value;
			}
		}

		
		public bool Init() {
			G.SaveRec = new List<SaveRecord[]>(new SaveRecord[(byte) RecType.Count][]);

			mnuMap = new MenuItem("Map Plug");
			mnuMap.MenuItems.Add("Location", new EventHandler(OnLocationWin));
			mnuSub = new MenuItem("Sub Windows");
			mnuMap.MenuItems.Add(mnuSub);
			mnuSub.MenuItems.Add("Properties", new EventHandler(OnLocPropWin));
			mnuSub.MenuItems.Add("L12 Tiles", new EventHandler(OnLayer12Swatch));
			mnuSub.MenuItems.Add("L3 Tiles", new EventHandler(OnLayer3Swatch));

			G.MainForm = (Form) G.DockMan.Parent;
			G.LocPropForm = new LocationPropertiesForm();

			G.L12TilesForm = new TilesForm();
			PictureBox imgTiles = G.L12TilesForm.GetTiles();
			imgTiles.MouseUp += new MouseEventHandler(G.L12TilesForm.OnMouseUp);
			imgTiles.MouseDown += new MouseEventHandler(G.L12TilesForm.OnMouseDown);
			imgTiles.MouseMove += new MouseEventHandler(G.L12TilesForm.OnMouseMove);
			G.L12TilesForm.nTileType = 0;
			G.L12TilesForm.TabText = "L12 Tiles";

			G.L3TilesForm = new TilesForm();
			imgTiles = G.L3TilesForm.GetTiles();
			imgTiles.MouseUp += new MouseEventHandler(G.L3TilesForm.OnMouseUp);
			imgTiles.MouseDown += new MouseEventHandler(G.L3TilesForm.OnMouseDown);
			imgTiles.MouseMove += new MouseEventHandler(G.L3TilesForm.OnMouseMove);
			G.L3TilesForm.nTileType = 1;
			G.L3TilesForm.TabText = "L3 Tiles";

			G.Init();

			return true;
		}


		public bool GetRecords() {
			G.PostStatus("Decompressing Location maps");
			#region Get Records
			G.SaveRec[(byte) RecType.LocMap] = new PlugRecord[0x100];
			ushort i;
			PlugRecord PlugRec;
			// This loop is the same code that's in GetSimpleRec, which could still be used here since
			// we're not doing anything different for PlugRecords
			for(i = 0; i < G.SaveRec[(byte) RecType.LocMap].Length; i++) {
				G.SaveRec[(byte) RecType.LocMap][i] = new PlugRecord();
				PlugRec = (PlugRecord) G.SaveRec[(byte) RecType.LocMap][i];
				PlugRec.nMaxSize = 0x6006;
				PlugRec.Pointer = new PointerRecord[1];
				PlugRec.Pointer[0] = new PointerRecord(G.GetRomAddr(PlugRomAddr.LocMap) + (i * 3));
				PlugRec.bCompressed = true;
				PlugRec.bCreateEmpty = false;
				PlugRec.Get();
			}

			G.PostStatus("Decompressing Location Tile Assembly");
			G.SaveRec[(byte) RecType.LocTileAsmL12] = new SaveRecord[0x57];
			G.GetSimpleRec(G.SaveRec[(byte) RecType.LocTileAsmL12], 0x1000, PlugRomAddr.LocTileAsmL12, false);

			G.SaveRec[(byte) RecType.LocTileAsmL3] = new SaveRecord[0x17];
			G.GetSimpleRec(G.SaveRec[(byte) RecType.LocTileAsmL3], 0x1000, PlugRomAddr.LocTileAsmL3, false);

			SaveRecord Rec;
			G.SaveRec[(byte) RecType.LocProp] = new SaveRecord[0x200];
			for(i = 0; i < G.SaveRec[(byte) RecType.LocProp].Length; i++) {
				G.SaveRec[(byte) RecType.LocProp][i] = new SaveRecord();
				Rec = G.SaveRec[(byte) RecType.LocProp][i];
				Rec.nMaxSize = 0x0E;
				Rec.nOrigSize = 0x0E;
				Rec.nOrigAddr = (uint) (G.GetRomAddr(PlugRomAddr.LocProp) + (i * 0x0E));
				Rec.bCompressed = false;
				Rec.bCreateEmpty = false;
				Rec.Get();
			}

			G.SaveRec[(byte) RecType.LocTileset] = new SaveRecord[0x40];
			for(i = 0; i < G.SaveRec[(byte) RecType.LocTileset].Length; i++) {
				G.SaveRec[(byte) RecType.LocTileset][i] = new SaveRecord();
				Rec = G.SaveRec[(byte) RecType.LocTileset][i];
				Rec.nMaxSize = 0x08;
				Rec.nOrigSize = 0x08;
				Rec.nOrigAddr = (uint) (G.GetRomAddr(PlugRomAddr.LocTileset) + (i << 3));
				Rec.bCompressed = false;
				Rec.bCreateEmpty = false;
				Rec.Get();
			}

			G.SaveRec[(byte) RecType.LocPal] = new SaveRecord[0x59];
			for(i = 0; i < G.SaveRec[(byte) RecType.LocPal].Length; i++) {
				G.SaveRec[(byte) RecType.LocPal][i] = new SaveRecord();
				Rec = G.SaveRec[(byte) RecType.LocPal][i];
				Rec.nMaxSize = 0xD2;
				Rec.nOrigSize = 0xD2;
				Rec.nOrigAddr = (uint) (G.GetRomAddr(PlugRomAddr.LocPal) + (i * 0xD2));
				Rec.bCompressed = false;
				Rec.bCreateEmpty = false;
				Rec.Get();
			}

			G.SaveRec[(byte) RecType.LocTilesL12] = new SaveRecord[0xE0];
			G.GetSimpleRec(G.SaveRec[(byte) RecType.LocTilesL12], 0x2000, PlugRomAddr.LocL12TilesetPtr, false);
			#endregion

			#region Data-dependant form setup
			int nIndex = G.LocPropForm.LocationCombo.SelectedIndex;
			if(nIndex == -1) {
				nIndex = 0;
			}
			G.LocPropForm.LocationCombo.Items.Clear();
			G.LocPropForm.LocationCombo.Items.AddRange(G.sLocation);
			LocationForm LocForm = G.LocPropForm.ActiveMap;
			G.LocPropForm.ActiveMap = null;
			G.LocPropForm.LocationCombo.SelectedIndex = nIndex;
			G.LocPropForm.ActiveMap = LocForm;
			#endregion

			return true;
		}


		public bool Close() {
			return true;
		}



		public void OnLocationWin(object sender, System.EventArgs e) {
			OnLocPropWin(sender, e);
			OnLayer12Swatch(sender, e);
			OnLayer3Swatch(sender, e);
			G.L12TilesForm.Activate();
			LocationForm LocForm = new LocationForm();
			LocForm.TabText = "Location - " + G.LocPropForm.LocationCombo.SelectedItem;
			LocForm.bNoUpdate = true;
			LocForm.MdiParent = G.MainForm;
			LocForm.Show(G.DockMan);
			LocForm.BringToFront();
			LocForm.bNoUpdate = false;

			G.LocPropForm.OnChangeLocation(null, EventArgs.Empty);
		}



		private void OnLocPropWin(object sender, System.EventArgs e) {
			G.LocPropForm.Show(G.DockMan);
		}



		private void OnLayer12Swatch(object sender, System.EventArgs e) {
			G.L12TilesForm.Show(G.DockMan);
		}



		private void OnLayer3Swatch(object sender, System.EventArgs e) {
			G.L3TilesForm.Show(G.DockMan);
		}
	}
}
