using System;
using System.Drawing;
using FluxShared;



namespace MapPlugin{
	public class cLocMap : cTileMap{
		public cLocMap(MapForm CallingForm) : base(CallingForm, 0){
		}



		public override void DrawSolidity() {
			Graphics graphMap = Graphics.FromImage(LocForm.imgMap.Image);
			graphMap.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceOver;
			SolidBrush SolidityBrush = new SolidBrush(Color.FromArgb(0x60, 0xFF, 0, 0));
			SolidBrush[] ZPlaneBrush = new SolidBrush[5];
			ZPlaneBrush[0] = new SolidBrush(Color.FromArgb(0xA0, 0xFF, 0, 0xFF));
			ZPlaneBrush[1] = new SolidBrush(Color.FromArgb(0xA0, 0, 0, 0xFF));
			ZPlaneBrush[2] = new SolidBrush(Color.FromArgb(0xA0, 0, 0xFF, 0xFF));
			ZPlaneBrush[3] = new SolidBrush(Color.FromArgb(0xA0, 0x80, 0x80, 0xFF));
			ZPlaneBrush[4] = new SolidBrush(Color.FromArgb(0xA0, 0xFF, 0xFF, 0));
			Pen StairsPen = new Pen(Color.FromArgb(0x60, 0, 0xFF, 00));
			SolidBrush BattleBrush = new SolidBrush(Color.FromArgb(0xA0, 0x80, 0xFF, 0x80));
			SolidBrush NPCBrush = new SolidBrush(Color.FromArgb(0xA0, 0xFF, 0x80, 0x80));
			Pen WindPen = new Pen(Color.FromArgb(0xA0, 0xFF, 0xFF, 0xFF));
			SolidBrush WindBrush = new SolidBrush(Color.FromArgb(0x60, 0x00, 0x00, 0x00));
			
			for(byte y = 0; y < nMaxHeight; y++){
				for(byte x = 0; x < nMaxWidth; x++){
					uint nTilePropSwatch = G.GetInt24(MapTileProp, 3 * ((y * nMaxWidth) + x));
					uint nSolidity = (nTilePropSwatch & 0x0000FC) >> 2;
					// Full Solidity
					if(LocForm.mnuSolidity.Checked){
						// Rounded Corners
						if(nSolidity > 0x01 && nSolidity < 0x16) {
							nSolidity -= 2;
							byte nSolidCorner = (byte) (nSolidity & 0x03);
							nSolidity >>= 2;

							byte nMod2 = 4;
							byte nMod3 = 16;
							byte nMod4 = 8;
							byte nNumPoints = 3;

							switch(nSolidity) {
							// 45%?
							case 0:
							default:
								if((nTilePropSwatch & 0x090000) == 0x080000) {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 16, 16);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								break;
							// 30%
							case 1:
								if((nTilePropSwatch & 0x090000) == 0x080000) {
									nSolidCorner = (byte) (3 - nSolidCorner);
									nNumPoints = 4;
									nMod4 = 5;
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									nMod3 = 11;
								}
								break;
							// 22%
							case 2:
								if((nTilePropSwatch & 0x090000) == 0x080000) {
									nSolidCorner = (byte) (3 - nSolidCorner);
									nNumPoints = 4;
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									nMod2 = 3;
								}
								break;
							// 75%
							case 3:
							case 4:
								if((nTilePropSwatch & 0x090000) == 0x080000) {
									nMod2 = 3;
									nMod3 = 8;
									nSolidCorner = (byte) (3 - nSolidCorner);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									nNumPoints = 5;
								}
								break;
							}

							Point p1 = new Point((x + (1 - (nSolidCorner & 0x01))) << 4, (y + (1 - ((nSolidCorner & 0x02) >> 1))) << 4);
							Point p2 = new Point(p1.X, p1.Y + (((nSolidCorner & 0x02) - 1) << nMod2));
							Point p3 = new Point(p1.X + ((((nSolidCorner & 0x01) << 1) - 1) * nMod3), p1.Y);

							if(nNumPoints == 3) {
								graphMap.FillPolygon(SolidityBrush, new Point[] { p1, p2, p3 });
							}
							else {
								Point p4 = new Point(p1.X + ((((nSolidCorner & 0x01) << 1) - 1) * nMod4), p1.Y + (((nSolidCorner & 0x02) - 1) << 4));
								if(nNumPoints == 4) {
									graphMap.FillPolygon(SolidityBrush, new Point[] { p1, p2, p4, p3 });
								}
								else {
									Point p5 = new Point(p1.X + ((((nSolidCorner & 0x01) << 1) - 1) << 4), p1.Y + (((nSolidCorner & 0x02) - 1) << 3));
									graphMap.FillPolygon(SolidityBrush, new Point[] { p1, p2, p4, p5, p3 });
								}
							}
						}
						else {
							switch(nSolidity) {
							// Walkable
							case 0x00:
							default:
								// Do nothing
								break;
							case 0x01:
								if((nTilePropSwatch & 0x280000) == 0) {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 16, 16);
								}
								break;
							// Stairs
							case 0x16:
								graphMap.DrawLine(StairsPen, x << 4, (y << 4) + 11, x << 4, (y << 4) + 15);
								graphMap.DrawLine(StairsPen, x << 4, (y << 4) + 15, (x << 4) + 3, (y << 4) + 15);
								graphMap.DrawLine(StairsPen, x << 4, (y << 4) + 15, (x << 4) + 15, y << 4);
								graphMap.DrawLine(StairsPen, (x << 4) + 11, y << 4, (x << 4) + 15, y << 4);
								graphMap.DrawLine(StairsPen, (x << 4) + 15, y << 4, (x << 4) + 15, (y << 4) + 3);
								break;
							case 0x17:
								graphMap.DrawLine(StairsPen, x << 4, (y << 4) + 3, x << 4, y << 4);
								graphMap.DrawLine(StairsPen, x << 4, y << 4, (x << 4) + 3, y << 4);
								graphMap.DrawLine(StairsPen, x << 4, y << 4, (x << 4) + 15, (y << 4) + 15);
								graphMap.DrawLine(StairsPen, (x << 4) + 11, (y << 4) + 15, (x << 4) + 15, (y << 4) + 15);
								graphMap.DrawLine(StairsPen, (x << 4) + 15, (y << 4) + 15, (x << 4) + 15, (y << 4) + 11);
								break;
							// Quads
							case 0x18:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 8, 16);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, y << 4, 8, 16);
								}
								break;
							case 0x19:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 16, 8);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, x << 4, (y << 4) + 8, 16, 8);
								}
								break;
							case 0x1A:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, x << 4, (y << 4) + 8, 8, 8);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 8, 8);
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, y << 4, 8, 16);
								}
								break;
							case 0x1B:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, (y << 4) + 8, 8, 8);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 8, 16);
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, y << 4, 8, 8);
								}
								break;
							case 0x1C:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, y << 4, 8, 8);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 8, 16);
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, (y << 4) + 8, 8, 8);
								}
								break;
							case 0x1D:
								if((nTilePropSwatch & 0x080000) != 0x080000) {
									graphMap.FillRectangle(SolidityBrush, x << 4, y << 4, 8, 8);
								}
								else if((nTilePropSwatch & 0x090000) == 0x090000) {
									continue;
								}
								else {
									graphMap.FillRectangle(SolidityBrush, x << 4, (y << 4) + 8, 8, 8);
									graphMap.FillRectangle(SolidityBrush, (x << 4) + 8, y << 4, 8, 16);
								}
								break;
							// Ladder
							case 0x1E:
								graphMap.DrawLine(StairsPen, (x << 4) + 2, y << 4, x << 4, (y << 4) + 2);
								graphMap.DrawLine(StairsPen, (x << 4) + 2, y << 4, (x << 4) + 4, (y << 4) + 2);
								graphMap.DrawLine(StairsPen, (x << 4) + 2, y << 4, (x << 4) + 2, (y << 4) + 15);
								graphMap.DrawLine(StairsPen, (x << 4) + 13, (y << 4) + 15, (x << 4) + 15, (y << 4) + 13);
								graphMap.DrawLine(StairsPen, (x << 4) + 13, (y << 4) + 15, (x << 4) + 11, (y << 4) + 13);
								graphMap.DrawLine(StairsPen, (x << 4) + 13, (y << 4) + 15, (x << 4) + 13, y << 4);
								break;
							}
						}
					}

					if(LocForm.mnuWind.Checked){
						// Wind
						byte nWindSpeed = (byte) ((nTilePropSwatch & 0x000C00) >> 10);
						byte nWindDir = (byte) ((nTilePropSwatch & 0x000300) >> 8);
						for(int i = 0; i < nWindSpeed; i++) {
							float[] wx = new float[2];
							float[] wy = new float[2];
							wx[0] = (x << 4) + 7;
							wx[1] = (x << 4) + ((i + 1) << 1) + 6 - nWindSpeed;
							wy[0] = (y << 4) + 7;
							wy[1] = (y << 4) + ((i + 1) << 1) + 6 - nWindSpeed;

							switch(nWindDir){
							case 0:
								if(i == 0) {
									graphMap.FillRectangle(WindBrush, wx[0] - 3, wy[1] - 1, 0x07, 0x05 + ((nWindSpeed - 1) << 1));
								}
								graphMap.DrawLine(WindPen, wx[0], wy[1], wx[0] - 2, wy[1] + 2);
								graphMap.DrawLine(WindPen, wx[0], wy[1], wx[0] + 2, wy[1] + 2);
								break;
							case 1:
								if(i == 0) {
									graphMap.FillRectangle(WindBrush, wx[0] - 3, wy[1] - 1, 0x07, 0x05 + ((nWindSpeed - 1) << 1));
								}
								graphMap.DrawLine(WindPen, wx[0], wy[1] + 2, wx[0] - 2, wy[1]);
								graphMap.DrawLine(WindPen, wx[0], wy[1] + 2, wx[0] + 2, wy[1]);
								break;
							case 2:
								if(i == 0) {
									graphMap.FillRectangle(WindBrush, wx[1] - 1, wy[0] - 3, 0x05 + ((nWindSpeed - 1) << 1), 0x07);
								}
								graphMap.DrawLine(WindPen, wx[1], wy[0], wx[1] + 2, wy[0] - 2);
								graphMap.DrawLine(WindPen, wx[1], wy[0], wx[1] + 2, wy[0] + 2);
								break;
							default:
								if(i == 0) {
									graphMap.FillRectangle(WindBrush, wx[1] - 1, wy[0] - 3, 0x05 + ((nWindSpeed - 1) << 1), 0x07);
								}
								graphMap.DrawLine(WindPen, wx[1] + 2, wy[0], wx[1], wy[0] - 2);
								graphMap.DrawLine(WindPen, wx[1] + 2, wy[0], wx[1], wy[0] + 2);
								break;
							}
						}
					}

					if(LocForm.mnuZPlane.Checked){
						// Neutral
						if((nTilePropSwatch & 0x200000) == 0x200000){
							graphMap.FillRectangle(ZPlaneBrush[4], x << 4, y << 4, 16, 16);
						}
							// Non-neutral
						else{
							graphMap.FillRectangle(ZPlaneBrush[(nTilePropSwatch & 0x030000) >> 16], x << 4, y << 4, 16, 16);
						}
					}

					if(LocForm.mnuBattle.Checked){
						// Neutral
						if((nTilePropSwatch & 0x008000) == 0x008000){
							graphMap.FillRectangle(BattleBrush, x << 4, y << 4, 16, 16);
						}
					}

					if(LocForm.mnuNPC.Checked){
						if((nTilePropSwatch & 0x800000) == 0x800000){
							graphMap.FillRectangle(NPCBrush, x << 4, y << 4, 16, 16);
						}
					}
				}
			}
		}
	}
}
