using System;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using PSVRender;
using System.IO;
using FluxShared;
using WeifenLuo.WinFormsUI.Docking;

namespace MapPlugin{
	public abstract class cTileMap{
		protected MapForm LocForm;
		public byte[] nLWidth = new byte[3];
		public byte[] nLHeight = new byte[3];
		public ushort[] nLOffset = new ushort[3];
		public byte nMaxWidth;
		public byte nMaxHeight;
		public byte[] MapData;
		public Bitmap bmpMap;
		byte nType;
		byte MapType;
		NumericUpDown MapSpin;
		public bool bDrawLayer3;
		public bool bAreaSelectStarted = false;
		public ushort nCopyAreaSelectX = 0;
		public ushort nCopyAreaSelectW = 0;
		public ushort nCopyAreaSelectY = 0;
		public ushort nCopyAreaSelectH = 0;
		public ushort nPasteAreaSelectX = 0;
		public ushort nPasteAreaSelectW = 0;
		public ushort nPasteAreaSelectY = 0;
		public ushort nPasteAreaSelectH = 0;
		public ushort[][][] AreaSelected = new ushort[3][][];
		public bool bCopyFinished = true;
		public Rectangle[] CoordRect;

		public byte[] MapTileProp;
		public ushort MapTilePropOff;
		public bool[] bLSub = new bool[3];
		public bool[] bLMain = new bool[3];
		public bool bSubtractive;
		public bool bHalfColor;
		public bool[] bTranslucent = new bool[3];
		public byte[][][] AreaTileProp;



		public cTileMap(MapForm CallingForm, byte nTypeIn){
			LocForm = CallingForm;
			nType = nTypeIn;
			
			if(nType == 0){
				MapType = (byte) RecType.LocMap;
				MapSpin = G.LocPropForm.MapSpin;
			}
			//else{
			//    MapType = (byte) RecType.OWMap;
			//    MapSpin = MForm.OWPropForm.MapSpin;
			//}
		}
			
		
		
		public void ChangeMap(){
			LocForm.imgMap.Image = new Bitmap(8, 8);
			MapData = new byte[0x6006];
			MapTileProp = new byte[0x3000];
			Array.Copy(G.SaveRec[MapType][(int) MapSpin.Value].nData, 0, MapData, 0, G.SaveRec[MapType][(int) MapSpin.Value].nDataSize);

			if(nType == 0) {
				DecodeMapHeader(this);
			}
			//else {
			//    for(byte i = 0; i < 3; i++) {
			//        nLWidth[i] = 0x60;
			//        nLHeight[i] = 0x40;
			//        bLSub[i] = true;
			//        bLMain[i] = false;
			//    }
			//    nLOffset[0] = 0;
			//    nLOffset[1] = 0x1800;
			//    bHalfColor = false;
			//    bDrawLayer3 = false;
			//}

			if(nLWidth[0] > nLWidth[1]) {
				nMaxWidth = nLWidth[0];
			}
			else{
				nMaxWidth = nLWidth[1];
			}

			if(nLHeight[0] > nLHeight[1]){
				nMaxHeight = nLHeight[0];
			}
			else{
				nMaxHeight = nLHeight[1];
			}

			LocForm.imgMap.Width = ((nMaxWidth << 4) + 4) * G.nZoomFactor;
			LocForm.imgMap.Height = ((nMaxHeight << 4) + 4) * G.nZoomFactor;
			bmpMap = new Bitmap(nMaxWidth << 4, nMaxHeight << 4);
			
			if(nType == 0){
				MapTilePropOff = nLOffset[2];
				if(bDrawLayer3){
					MapTilePropOff += (ushort) (nLHeight[2] * nLWidth[2]);
				}

				G.PostStatus("Decoding extra map data");
				if(G.nRomType == 2) {
					uint nMapSize = (uint) (nMaxHeight * nMaxWidth);
					for(int i = 0; i < nMapSize; i++) {
						MapTileProp[i * 3] = MapData[i + MapTilePropOff];
						MapTileProp[(i * 3) + 1] = MapData[i + MapTilePropOff + nMapSize];
						MapTileProp[(i * 3) + 2] = MapData[i + MapTilePropOff + (nMapSize << 1)];
					}
				}
				else {
					CompressionDataVar CDV = new CompressionDataVar();
					CDV.nDecompressedSize = (uint) (3 * nMaxHeight * nMaxWidth);
					CDV.SrcBuffer = MapData;
					CDV.WorkingBuffer = MapTileProp;
					CDV.nSrcOff = MapTilePropOff;
					CDV.nDecRoutine = CompressionRoutines.eChronoTriggerRunLength;
					CDV.nWorkOff = 0;
					CDV.DecompressData();
				}
				Array.Clear(MapData, MapTilePropOff, 0x6006 - MapTilePropOff);
			}
		}



		public void DecodeMapHeader(cTileMap TileMapIn) {
			nLOffset[0] = 6;
			nLWidth[0] = (byte) (((MapData[0] & 0x03) + 1) << 4);
			nLHeight[0] = (byte) (((MapData[0] & 0x0C) + 4) << 2);
			nLOffset[1] = (ushort) (nLOffset[0] + (nLWidth[0] * nLHeight[0]));
			nLWidth[1] = (byte) ((MapData[0] & 0x30) + 0x10);
			nLHeight[1] = (byte) (((MapData[0] & 0xC0) + 0x40) >> 2);
			nLOffset[2] = (ushort) (nLOffset[1] + (nLWidth[1] * nLHeight[1]));
			nLWidth[2] = (byte) (((MapData[1] & 0x03) + 1) << 4);
			nLHeight[2] = (byte) (((MapData[1] & 0x0C) + 4) << 2);

			bDrawLayer3 = Convert.ToBoolean(MapData[1] & 0x80);

			byte i;
			for(i = 0; i < 3; i++) {
				bLMain[i] = Convert.ToBoolean(MapData[4] & (1 << i));
				bLSub[i] = Convert.ToBoolean(MapData[4] & (1 << (4 | i)));
			}

			bSubtractive = Convert.ToBoolean((MapData[5] & 0x80) >> 7);
			bHalfColor = Convert.ToBoolean((MapData[5] & 0x40) >> 6);

			for(i = 0; i < 3; i++) {
				bTranslucent[i] = Convert.ToBoolean(MapData[5] & (1 << i));
			}
		}



		public void ChangeMapSize(){
			LocForm.imgMap.Image = new Bitmap(8, 8);

			byte[] NewMapData = new byte[0x6006];
			byte[] NewMapTileProp = new byte[0x3000];

			Array.Copy(MapData, 0, NewMapData, 0, 6);

			byte nNewL1Width = (byte) (((MapData[0] & 0x03) + 1) << 4);
			byte nNewL1Height = (byte) (((MapData[0] & 0x0C) + 4) << 2);
			byte nNewL2Width = (byte) ((MapData[0] & 0x30) + 0x10);
			byte nNewL2Height = (byte) (((MapData[0] & 0xC0) + 0x40) >> 2);
			byte nNewL3Width = (byte) (((MapData[1] & 0x03) + 1) << 4);
			byte nNewL3Height = (byte) (((MapData[1] & 0x0C) + 4) << 2);
			byte bNewDrawLayer3 = (byte) ((MapData[1] & 0x80) >> 7);

			byte nNewL1Offset = 6;
			ushort nNewL2Offset = (ushort) (nNewL1Offset + (nNewL1Width * nNewL1Height));
			ushort nNewL3Offset = (ushort) (nNewL2Offset + (nNewL2Width * nNewL2Height));
			ushort nNewTilePropOff = nNewL3Offset;
			if(bNewDrawLayer3 == 1){
				nNewTilePropOff += (ushort) (nNewL3Width * nNewL3Height);
			}

			byte nL1X = nNewL1Width;
			if(nLWidth[0] < nNewL1Width){
				nL1X = nLWidth[0];
			}
			byte nL1Y = nNewL1Height;
			if(nLHeight[0] < nNewL1Height){
				nL1Y = nLHeight[0];
			}
			byte nL2X = nNewL2Width;
			if(nLWidth[1] < nNewL2Width){
				nL2X = nLWidth[1];
			}
			byte nL2Y = nNewL2Height;
			if(nLHeight[1] < nNewL2Height){
				nL2Y = nLHeight[1];
			}
			byte nL3X = nNewL3Width;
			if(nLWidth[2] < nNewL3Width){
				nL3X = nLWidth[2];
			}
			byte nL3Y = nNewL3Height;
			if(nLHeight[2] < nNewL3Height){
				nL3Y = nLHeight[2];
			}
			byte nNewMaxWidth = nNewL2Width;
			if(nNewL1Width > nNewL2Width){
				nNewMaxWidth = nNewL1Width;
			}
			byte nNewMaxHeight = nNewL2Height;
			if(nNewL1Height > nNewL2Height){
				nNewMaxHeight = nNewL1Height;
			}

			short x;
			sbyte y;
			for(y = 0; y < nL1Y; y++){
				for(x = 0; x < nL1X; x++){
					NewMapData[nNewL1Offset + (y * nNewL1Width) + x] = MapData[nLOffset[0] + (y * nLWidth[0]) + x];
				}
			}
			for(y = 0; y < nL2Y; y++){
				for(x = 0; x < nL2X; x++){
					NewMapData[nNewL2Offset + (y * nNewL2Width) + x] = MapData[nLOffset[1] + (y * nLWidth[1]) + x];
				}
			}
			if(bNewDrawLayer3 == 1 && bDrawLayer3){
				for(y = 0; y < nL3Y; y++){
					for(x = 0; x < nL3X; x++){
						NewMapData[nNewL3Offset + (y * nNewL3Width) + x] = MapData[nLOffset[2] + (y * nLWidth[2]) + x];
					}
				}
			}

			MapData = NewMapData;

			nLWidth[0] = nNewL1Width;
			nLHeight[0] = nNewL1Height;
			nLWidth[1] = nNewL2Width;
			nLHeight[1] = nNewL2Height;
			nLWidth[2] = nNewL3Width;
			nLHeight[2] = nNewL3Height;

			short nMaxX = (short) nNewMaxWidth;
			sbyte nMaxY = (sbyte) nNewMaxHeight;

			if(nMaxWidth < nNewMaxWidth){
				nMaxX = (short) nMaxWidth;
			}
			if(nMaxHeight < nNewMaxHeight){
				nMaxY = (sbyte) nMaxHeight;
			}

			for(y = 0; y < nMaxY; y++){
				for(x = 0; x < (nMaxX * 3); x += 3){
					NewMapTileProp[(y * nNewMaxWidth * 3) + x] = MapTileProp[(y * nMaxWidth * 3) + x];
					NewMapTileProp[(y * nNewMaxWidth * 3) + x + 1] = MapTileProp[(y * nMaxWidth * 3) + x + 1];
					NewMapTileProp[(y * nNewMaxWidth * 3) + x + 2] = MapTileProp[(y * nMaxWidth * 3) + x + 2];
				}
			}

			MapTileProp = NewMapTileProp;

			nMaxWidth = nNewMaxWidth;
			nMaxHeight = nNewMaxHeight;

			LocForm.imgMap.Width = ((nMaxWidth << 4) + 4) * G.nZoomFactor;
			LocForm.imgMap.Height = ((nMaxHeight << 4) + 4) * G.nZoomFactor;
			bmpMap = new Bitmap(nMaxWidth << 4, nMaxHeight << 4);
			
			nLOffset[0] = nNewL1Offset;
			nLOffset[1] = nNewL2Offset;
			nLOffset[2] = nNewL3Offset;
			MapTilePropOff = nNewTilePropOff;

			bDrawLayer3 = Convert.ToBoolean(bNewDrawLayer3);

			RedrawMap();
		}



		public void RedrawMap() {
			ChangeMapSection(0, nMaxHeight, 0, nMaxWidth);
		}



		unsafe public void ChangeMapSection(byte nTop, byte nBottom, byte nLeft, byte nRight){
			byte yt;
			byte xt;
			ushort[] nTile = new ushort[3];
			bool[] bLInbounds = new bool[3];
			bool[] bLPrior = new bool[3];
			sbyte i;
			bool[] bDrawLayer = new bool[3];

			G.PostStatus("Drawing tiles");

			// Check bitmap boundaries
			if(nMaxWidth < nRight){
				nRight = nMaxWidth;
			}
			if(nMaxHeight < nBottom){
				nBottom = nMaxHeight;
			}

			// Lock Bitmap for pixel work
			BitmapData BMData = bmpMap.LockBits(new Rectangle(0, 0, bmpMap.Width, bmpMap.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);

			PutSubtileVar[] LSubtile = new PutSubtileVar[3];
			for(i = 0; i < 2; i++){
				LSubtile[i] = new PutSubtileVar();
				LSubtile[i].nBitPlanes = BitPlanes.eTetraPlane;
				LSubtile[i].WorkingBuffer = LocForm.L12SubtileData;
				LSubtile[i].WorkingImage = bmpMap;
				LSubtile[i].WorkingPalette = LocForm.LPalette;
				LSubtile[i].bHalfColorRender = bHalfColor;
				LSubtile[i].nWorkingImage = (byte*) BMData.Scan0.ToPointer();
			}

			if(nType == 0){
				LSubtile[2] = new PutSubtileVar();
				LSubtile[2].nBitPlanes = BitPlanes.eDiPlane;
				LSubtile[2].WorkingBuffer = LocForm.L3SubtileData;
				LSubtile[2].WorkingImage = bmpMap;
				LSubtile[2].WorkingPalette = LocForm.LPalette;
				LSubtile[2].bHalfColorRender = bHalfColor;
				LSubtile[2].nWorkingImage = (byte*) BMData.Scan0.ToPointer();
			}

			// Clear area
			LSubtile[0].ClearArea(nTop, nBottom, nLeft, nRight);

			// Rendering Order (only valid if using screen mode 1
			// (front most layers)
			// main screens
			//      bg 3, 1
			//		bg 1, 1
			//		bg 2, 1
			//		bg 1, 0
			//		bg 2, 0
			//		bg 3, 0
			// sub screens
			//      bg 3, 1
			//		bg 1, 1
			//		bg 2, 1
			//		bg 1, 0
			//		bg 2, 0
			//		bg 3, 0
			//(back most layers)

			bDrawLayer[2] = LocForm.mnuLayer[2].Checked && bDrawLayer3;
			bDrawLayer[1] = LocForm.mnuLayer[1].Checked;
			bDrawLayer[0] = LocForm.mnuLayer[0].Checked;

			byte[] nLayerOrder = { 2, 1, 0, 1, 0, 2 };
			// Subscreens
			for(i = 0; i < 6; i++){
				if(bDrawLayer[nLayerOrder[i]] && bLSub[nLayerOrder[i]]){
					for(yt = nTop; yt < nBottom; yt++){
						for(xt = nLeft; xt < nRight; xt++){
							byte nChunk;
							ushort nTilePropOff = (ushort) (3 * ((yt * nMaxWidth) + xt));

							bLInbounds[nLayerOrder[i]] = nLWidth[nLayerOrder[i]] > xt && nLHeight[nLayerOrder[i]] > yt;
					
							if(nType == 0){
								if(bLInbounds[nLayerOrder[i]]){
									if(nLayerOrder[i] == 2){
										nTile[2] = (ushort) (0x200 | MapData[nLOffset[2] + (yt * nLWidth[2]) + xt]);
									}
									else if(nLayerOrder[i] == 1){
										nTile[1] = (ushort) (((MapTileProp[nTilePropOff] & 0x02) << 7) | MapData[nLOffset[1] + (yt * nLWidth[1]) + xt]);
									}
									else{
										nTile[0] = (ushort) (((MapTileProp[nTilePropOff] & 0x01) << 8) | MapData[6 + (yt * nLWidth[0]) + xt]);
									}
								}
							}
							else{
								if(bLInbounds[nLayerOrder[i]]){
									if(nLayerOrder[i] == 1){
										nTile[1] = (ushort) (MapData[nLOffset[1] + (yt * nLWidth[0]) + xt] | 0x0100);
									}
									else{
										nTile[0] = (ushort) (MapData[(yt * nLWidth[0]) + xt]);
									}
								}
							}

							// Unbelievably, this is actually faster than getting a pixel from the already rendered bitmaps or even stored pixel arrays
							for(nChunk = 0; nChunk < 4; nChunk++){
								bLPrior[nLayerOrder[i]] = LocForm.nTilePrior[(nTile[nLayerOrder[i]] << 2) | nChunk] == 1;
								if(i < 3){
									if(bLInbounds[nLayerOrder[i]] && !bLPrior[nLayerOrder[i]]){
										RenderSubtile(nTile[nLayerOrder[i]], nChunk, LSubtile[nLayerOrder[i]], xt, yt);
									}
								}
								else{
									if(bDrawLayer[nLayerOrder[i]] && bLPrior[nLayerOrder[i]]){
										RenderSubtile(nTile[nLayerOrder[i]], nChunk, LSubtile[nLayerOrder[i]], xt, yt);
									}
								}
							}
						}
					}
				}
			}

			// Mainscreen rendering must be done completely separate from subscreen rendering
			// in order to get transparency effects
			if(nType == 0){
				// Prepare for Mainscreen rendering
				bmpMap.UnlockBits(BMData);
				Bitmap bmpSubmap = (Bitmap) bmpMap.Clone(new Rectangle(0, 0, bmpMap.Width, bmpMap.Height), PixelFormat.Format32bppArgb);
				BMData = bmpMap.LockBits(new Rectangle(0, 0, bmpMap.Width, bmpMap.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
				BitmapData BSData = bmpSubmap.LockBits(new Rectangle(0, 0, bmpSubmap.Width, bmpSubmap.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
				for(i = 0; i < 3; i++){
					LSubtile[i].nWorkingImage = (byte*) BMData.Scan0.ToPointer();
					LSubtile[i].nSubImage = (byte*) BSData.Scan0.ToPointer();

					if(bTranslucent[i]){
						if(bSubtractive){
							LSubtile[i].PRMode = PixelRenderMode.eSubtractive;
						}
						else{
							LSubtile[i].PRMode = PixelRenderMode.eAdditive;
						}
					}
				}

				// Mainscreens
				for(i = 0; i < 6; i++){
					if(bDrawLayer[nLayerOrder[i]] && bLMain[nLayerOrder[i]]){
						for(yt = nTop; yt < nBottom; yt++){
							for(xt = nLeft; xt < nRight; xt++){
								byte nChunk;
								ushort nTilePropOff = (ushort) (3 * ((yt * nMaxWidth) + xt));

								bLInbounds[nLayerOrder[i]] = nLWidth[nLayerOrder[i]] > xt && nLHeight[nLayerOrder[i]] > yt;
					
								if(bLInbounds[nLayerOrder[i]]){
									if(nType == 0){
										if(nLayerOrder[i] == 2){
											nTile[2] = (ushort) (0x200 | MapData[nLOffset[2] + (yt * nLWidth[2]) + xt]);
										}
										else if(nLayerOrder[i] == 1){
											nTile[1] = (ushort) (((MapTileProp[nTilePropOff] & 0x02) << 7) | MapData[nLOffset[1] + (yt * nLWidth[1]) + xt]);
										}
										else{
											nTile[0] = (ushort) (((MapTileProp[nTilePropOff] & 0x01) << 8) | MapData[6 + (yt * nLWidth[0]) + xt]);
										}
									}
									else{
										if(nLayerOrder[i] == 1){
											nTile[1] = (ushort) (MapData[nLOffset[1] + (yt * nLWidth[0]) + xt] | 0x0100);
										}
										else{
											nTile[0] = (ushort) (MapData[(yt * nLWidth[0]) + xt]);
										}
									}

									// Unbelievably, this is actually faster than getting a pixel from the already rendered bitmaps or even stored pixel arrays
									for(nChunk = 0; nChunk < 4; nChunk++){
										bLPrior[nLayerOrder[i]] = LocForm.nTilePrior[(nTile[nLayerOrder[i]] << 2) | nChunk] == 1;
										if(i < 3){
											if(!bLPrior[nLayerOrder[i]]){
												RenderSubtile(nTile[nLayerOrder[i]], nChunk, LSubtile[nLayerOrder[i]], xt, yt);
											}
										}
										else{
											if(bLPrior[nLayerOrder[i]]){
												RenderSubtile(nTile[nLayerOrder[i]], nChunk, LSubtile[nLayerOrder[i]], xt, yt);
											}
										}
									}
								}
							}
						}
					}
				}
				bmpSubmap.UnlockBits(BSData);
			}
			
			bmpMap.UnlockBits(BMData);

			RestoreMap();
			G.PostStatus("Finished Rendering");
		}



		void RenderSubtile(ushort nTile, byte nChunk, PutSubtileVar TileChunk, byte xt, byte yt){
			ushort nSubtile;
			if((nTile & 0x200) != 0x200){
				nSubtile = G.GetShort(LocForm.L12TileAsm, (nTile << 3) + (nChunk << 1));
				if(nType == 0){
					TileChunk.nDecodingStart = (ushort) (((nSubtile & 0x03FF) - 0x0100) << 5);
				}
				else{
					TileChunk.nDecodingStart = (ushort) ((nSubtile & 0x03FF) << 5);
				}
				if(TileChunk.nDecodingStart < 0x6000 + (nType * 0x1000)){
					TileChunk.nFlip = (byte) (nSubtile >> 14);
					TileChunk.nPaletteShift = (byte) ((nSubtile & 0x1C00) >> 10);
					TileChunk.dx = (int) ((uint) (xt << 4) | (uint) ((nChunk & 0x01) << 3));
					TileChunk.dy = (int) ((uint) (yt << 4) | (uint) ((nChunk & 0x02) << 2));
					TileChunk.PutSubtile();
				}
			}
			else if(LocForm.L3TileAsm != null){
				nTile &= 0xFF;
				nSubtile = G.GetShort(LocForm.L3TileAsm, (nTile << 3) + (nChunk << 1));
				TileChunk.nDecodingStart = (ushort) ((nSubtile & 0x03FF) << 4);
				if(TileChunk.nDecodingStart < 0x1000){
					TileChunk.nFlip = (byte) (nSubtile >> 14);
					TileChunk.nPaletteShift = (byte) ((nSubtile & 0x1C00) >> 10);
					TileChunk.dx = (int) ((uint) (xt << 4) | (uint) ((nChunk & 0x01) << 3));
					TileChunk.dy = (int) ((uint) (yt << 4) | (uint) ((nChunk & 0x02) << 2));
					TileChunk.PutSubtile();
				}
			}
		}

		
		
		public void SelectMapTile(MouseEventArgs mea){
			bAreaSelectStarted = false;

			ushort nMouseTileX = (ushort) (mea.X & 0xFFF0);
			ushort nMouseTileY = (ushort) (mea.Y & 0xFFF0);

			if(nMouseTileX < nCopyAreaSelectX){
				nCopyAreaSelectX = nMouseTileX;
			}
			if(nMouseTileY < nCopyAreaSelectY){
				nCopyAreaSelectY = nMouseTileY;
			}

			nCopyAreaSelectY >>= 4;
			nCopyAreaSelectX >>= 4;
			nCopyAreaSelectH >>= 4;
			nCopyAreaSelectW >>= 4;

			if(nType == 0){
				// Set recent map
				G.RecentMap = this;
			}

			if(nCopyAreaSelectY + nCopyAreaSelectH > nMaxHeight){
				nCopyAreaSelectH = (ushort) (nMaxHeight - nCopyAreaSelectY);
			}
			if(nCopyAreaSelectX + nCopyAreaSelectW > nMaxWidth){
				nCopyAreaSelectW = (ushort) (nMaxWidth - nCopyAreaSelectX);
			}
					
			if(!(nCopyAreaSelectH == 0 || nCopyAreaSelectH > nMaxHeight || nCopyAreaSelectW == 0 || nCopyAreaSelectW > nMaxWidth)){
				byte i;
				byte j;
				for(i = 0; i < 3; i++) {
					AreaSelected[i] = new ushort[nCopyAreaSelectW][];
					for(j = 0; j < nCopyAreaSelectW; j++) {
						AreaSelected[i][j] = new ushort[nCopyAreaSelectH];
					}
				}

				AreaTileProp = new byte[nCopyAreaSelectW][][];
				for(i = 0; i < nCopyAreaSelectW; i++){
					AreaTileProp[i] = new byte[nCopyAreaSelectH][];
					for(j = 0; j < nCopyAreaSelectH; j++){
						AreaTileProp[i][j] = new byte[3];
					}
				}
					
				for(byte y = 0; y < nCopyAreaSelectH; y++){
					for(byte x = 0; x < nCopyAreaSelectW; x++){
						uint nIndex = (uint) (3 * (((y + nCopyAreaSelectY) * nMaxWidth) + x + nCopyAreaSelectX));
						// Copies for only one map type
						if(nType == 0){
							AreaTileProp[x][y][0] = MapTileProp[nIndex];
							AreaTileProp[x][y][1] = MapTileProp[nIndex + 1];
							AreaTileProp[x][y][2] = MapTileProp[nIndex + 2];
							AreaSelected[2][x][y] = MapData[nLOffset[2] + ((y + nCopyAreaSelectY) * nLWidth[2]) + x + nCopyAreaSelectX];
						}
						//else if(nType == 1){
						//    byte nTransIndex = G.SaveRec[(byte) RecType.OWMTD][(int) MForm.OWPropForm.MusicTransitionSpin.Value].nData[(((y + nCopyAreaSelectY) * nMaxWidth) + x + nCopyAreaSelectX) >> 1];
						//    if(((x + nCopyAreaSelectX) & 0x01) == 0x01){
						//        nTransIndex &= 0x07;
						//    }
						//    else{
						//        nTransIndex >>= 4;
						//    }
						//    AreaTileProp[x][y][0] = nTransIndex;
						//}

						// Copies for both map types
						if(nType == 0){
							AreaSelected[0][x][y] = (ushort) ((MapTileProp[nIndex] & 0x01) << 8);
						}
						AreaSelected[0][x][y] |= (ushort) (MapData[nLOffset[0] + ((y + nCopyAreaSelectY) * nLWidth[0]) + x + nCopyAreaSelectX]);
						if(nType == 0){
							AreaSelected[1][x][y] = (ushort) ((MapTileProp[nIndex] & 0x02) << 7);
						}
						//else{
						//    AreaSelected[1][x][y] = 0x0100;
						//}
						AreaSelected[1][x][y] |= (ushort) (MapData[nLOffset[1] + ((y + nCopyAreaSelectY) * nLWidth[1]) + x + nCopyAreaSelectX]);
					}
				}
			}
			RestoreMap();
		}



		public void ChangeMapTile(MouseEventArgs mea){
			if(bCopyFinished == false){
				return;
			}
			
			// Select where to copy from
			ushort nSrcH = nCopyAreaSelectH;
			ushort nSrcW = nCopyAreaSelectW;

			if(nType == 0 && LocForm.mnuExplicit[5].Checked) {
				nSrcH = G.RecentMap.nCopyAreaSelectH;
				nSrcW = G.RecentMap.nCopyAreaSelectW;
			}

			if(LocForm.mnuMusicTransition.Checked && AreaTileProp == null) {
				nSrcH = 1;
				nSrcW = 1;
			}
			if(!bAreaSelectStarted){
				if(!(nSrcH == 0 || nSrcH > nMaxHeight || nSrcW == 0 || nSrcW > nMaxWidth)){
					byte xt = (byte) (mea.X >> 4);
					byte yt = (byte) (mea.Y >> 4);
					for(byte dy = 0; dy < nSrcH; dy++){
						byte y = (byte) (dy + yt);
						if(y >= nMaxHeight){
							continue;
						}
						for(byte dx = 0; dx < nSrcW; dx++){
							byte x = (byte) (dx + xt);
							if(x >= nMaxWidth){
								continue;
							}
							CopyTile(dx, x, dy, y);
						}
					}
					
					ChangeMapSection(yt, (byte) (yt + nSrcH), xt, (byte) (xt + nSrcW));
				}
			}
			else{
				bAreaSelectStarted = false;

				ushort nMouseTileX = (ushort) (mea.X & 0xFFF0);
				ushort nMouseTileY = (ushort) (mea.Y & 0xFFF0);

				if(nMouseTileX < nPasteAreaSelectX){
					nPasteAreaSelectX = nMouseTileX;
				}
				if(nMouseTileY < nPasteAreaSelectY){
					nPasteAreaSelectY = nMouseTileY;
				}

				nPasteAreaSelectY >>= 4;
				nPasteAreaSelectX >>= 4;
				nPasteAreaSelectH >>= 4;
				nPasteAreaSelectW >>= 4;

				if(nPasteAreaSelectY + nPasteAreaSelectH > nMaxHeight){
					nPasteAreaSelectH = (ushort) (nMaxHeight - nPasteAreaSelectY);
				}
				if(nPasteAreaSelectX + nPasteAreaSelectW > nMaxWidth){
					nPasteAreaSelectW = (ushort) (nMaxWidth - nPasteAreaSelectX);
				}

				if(!(nPasteAreaSelectH == 0 || nPasteAreaSelectH > nMaxHeight || nPasteAreaSelectW == 0 || nPasteAreaSelectW > nMaxWidth) && nSrcH != 0 && nSrcW != 0){
					byte xc;
					byte yc = 0;
					for(byte dy = 0; dy < nPasteAreaSelectH; dy++){
						xc = 0;
						for(byte dx = 0; dx < nPasteAreaSelectW; dx++){
							byte x = (byte) (dx + nPasteAreaSelectX);
							byte y = (byte) (dy + nPasteAreaSelectY);
							CopyTile(xc, x, yc, y);
							xc++;
							if(xc == nSrcW){
								xc = 0;
							}
						}
						yc++;
						if(yc == nSrcH){
							yc = 0;
						}
					}
					ChangeMapSection((byte) nPasteAreaSelectY, (byte) (nPasteAreaSelectY + nPasteAreaSelectH), (byte) nPasteAreaSelectX, (byte) (nPasteAreaSelectX + nPasteAreaSelectW));
				}
			}
		}



		private void CopyTile(byte dx, byte x, byte dy, byte y) {
			// Select where to copy from
			ushort[][][] SrcArea = AreaSelected;
			byte[][][] SrcProp = AreaTileProp;
			ushort nSrcH = nCopyAreaSelectH;
			ushort nSrcW = nCopyAreaSelectW;

			if(nType == 0 && LocForm.mnuExplicit[5].Checked) {
				SrcArea = G.RecentMap.AreaSelected;
				SrcProp = G.RecentMap.AreaTileProp;
				nSrcH = G.RecentMap.nCopyAreaSelectH;
				nSrcW = G.RecentMap.nCopyAreaSelectW;
			}

			if(LocForm.mnuExplicit[0].Checked) {
				if(nType == 0 && LocForm.mnuLayer[2].Checked) {
					// Testing layer size separate from other tests keeps tiles from pasting onto other layers
					if(nLWidth[2] > x && nLHeight[2] > y) {
						MapData[nLOffset[2] + (y * nLWidth[2]) + x] = (byte) SrcArea[2][dx][dy];
					}
				}
				else if(LocForm.mnuLayer[0].Checked) {
					if(nLWidth[0] > x && nLHeight[0] > y) {
						if(nType == 0) {
							MapTileProp[3 * ((y * nMaxWidth) + x)] &= 0x02;
							MapTileProp[3 * ((y * nMaxWidth) + x)] |= (byte) (SrcArea[0][dx][dy] >> 8);
						}
						MapData[nLOffset[0] + (y * nLWidth[0]) + x] = (byte) (SrcArea[0][dx][dy] & 0xFF);
					}
				}
				else if(LocForm.mnuLayer[1].Checked) {
					if(nLWidth[1] > x && nLHeight[1] > y) {
						if(nType == 0) {
							MapTileProp[3 * ((y * nMaxWidth) + x)] &= 0x01;
							MapTileProp[3 * ((y * nMaxWidth) + x)] |= (byte) ((SrcArea[1][dx][dy] >> 7) & 0x02);
							MapTileProp[3 * ((y * nMaxWidth) + x)] |= (byte) (SrcProp[dx][dy][0] & 0xFC);
							MapTileProp[(3 * ((y * nMaxWidth) + x)) + 1] = SrcProp[dx][dy][1];
							MapTileProp[(3 * ((y * nMaxWidth) + x)) + 2] = SrcProp[dx][dy][2];
						}
						MapData[nLOffset[1] + (y * nLWidth[1]) + x] = (byte) (SrcArea[1][dx][dy] & 0xFF);
					}
				}
			}
			else {
				if(LocForm.mnuExplicit[3].Checked) {
					// Testing layer size separate from other tests keeps tiles from pasting onto other layers
					if(nLWidth[2] > x && nLHeight[2] > y) {
						MapData[nLOffset[2] + (y * nLWidth[2]) + x] = (byte) SrcArea[2][dx][dy];
					}
				}
				byte nSrcLayer;
				if(LocForm.mnuExplicit[1].Checked) {
					nSrcLayer = Convert.ToByte(LocForm.mnuExplicit[6].Checked);
					if(nLWidth[0] > x && nLHeight[0] > y) {
						if(nType == 0) {
							MapTileProp[3 * ((y * nMaxWidth) + x)] &= 0xFE;
							MapTileProp[3 * ((y * nMaxWidth) + x)] |= (byte) (SrcArea[nSrcLayer][dx][dy] >> 8);
						}
						MapData[nLOffset[0] + (y * nLWidth[0]) + x] = (byte) (SrcArea[nSrcLayer][dx][dy] & 0xFF);
					}
				}
				if(LocForm.mnuExplicit[2].Checked) {
					nSrcLayer = Convert.ToByte(!LocForm.mnuExplicit[6].Checked);
					if(nLWidth[1] > x && nLHeight[1] > y) {
						if(nType == 0) {
							MapTileProp[3 * ((y * nMaxWidth) + x)] &= 0xFD;
							MapTileProp[3 * ((y * nMaxWidth) + x)] |= (byte) ((SrcArea[nSrcLayer][dx][dy] >> 7) & 0x02);
						}
						MapData[nLOffset[1] + (y * nLWidth[1]) + x] = (byte) (SrcArea[nSrcLayer][dx][dy] & 0xFF);
					}
				}
			}
		}




		public void RestoreMap(){
			LocForm.imgMap.Image = bmpMap.Clone(new Rectangle(0, 0, bmpMap.Width, bmpMap.Height), bmpMap.PixelFormat);
			if(LocForm.mnuScrollMask.Checked){
				DrawScrollBoundary();
			}
			if(LocForm.mnuSolidity.Checked || LocForm.mnuWind.Checked || LocForm.mnuZPlane.Checked || LocForm.mnuBattle.Checked || LocForm.mnuNPC.Checked){
				DrawSolidity();
			}
			if(LocForm.mnuDoorTrigger.Checked){
				DrawDoors();
			}
			if(LocForm.mnuUnk120.Checked){
				DrawUnknown();
			}
			DrawCoordinates();
			GC.Collect();
			LocForm.imgMap.Invalidate();
		}



		public void DrawScrollBoundary(){
			Graphics graphMap = Graphics.FromImage(LocForm.imgMap.Image);
			graphMap.CompositingMode = System.Drawing.Drawing2D.CompositingMode.SourceOver;
			SolidBrush ScrollBrush = new SolidBrush(Color.FromArgb(0x80, 0, 0, 0));
			int nTop = (int) G.LocPropForm.TopSpin.Value << 4;
			int nBottom = (int) G.LocPropForm.BottomSpin.Value << 4;
			int nRight = (int) G.LocPropForm.RightSpin.Value << 4;
			int nLeft = (int) G.LocPropForm.LeftSpin.Value << 4;
			if(nLeft == 0x800){
				nLeft = 0;
				nTop = 0;
				nRight = 0x3F0;
				nBottom = 0x3F0;
			}
			graphMap.FillRectangle(ScrollBrush, 0, 0, bmpMap.Width, nTop);
			graphMap.FillRectangle(ScrollBrush, nRight + 16, nTop, bmpMap.Width - (nRight + 16), bmpMap.Height - nTop);
			graphMap.FillRectangle(ScrollBrush, 0, nBottom + 16, nRight + 16, bmpMap.Height - (nBottom + 16));
			graphMap.FillRectangle(ScrollBrush, 0, nTop, nLeft, nBottom + 16 - nTop);
		}



		public abstract void DrawSolidity();



		public void DrawCoordinates() {
			if(CoordRect == null) {
				return;
			}

			Pen DrawBrush = new Pen(Color.FromArgb(0xFF, 0x80, 0x80, 0xFF), 1);
			Graphics graphMap = Graphics.FromImage(LocForm.imgMap.Image);

			for(ushort i = 0; i < CoordRect.Length; i++) {
				graphMap.DrawRectangle(DrawBrush, CoordRect[i]);
			}
		}



		void DrawDoors() {
			SolidBrush DoorBrush = new SolidBrush(Color.FromArgb(0x80, 0x80, 0xFF, 0xFF));
			Graphics graphMap = Graphics.FromImage(LocForm.imgMap.Image);

			for(byte y = 0; y < nMaxHeight; y++){
				for(byte x = 0; x < nMaxWidth; x++){
					uint nTilePropSwatch = G.GetInt24(MapTileProp, 3 * ((y * nMaxWidth) + x));
					if((nTilePropSwatch & 0x001000) == 0x001000){
						graphMap.FillRectangle(DoorBrush, x << 4, y << 4, 16, 16);
					}
				}
			}
		}



		void DrawUnknown(){
			SolidBrush UnkBrush = new SolidBrush(Color.FromArgb(0x80, 0xFF, 0x80, 0xFF));
			Graphics graphMap = Graphics.FromImage(LocForm.imgMap.Image);

			for(byte y = 0; y < nMaxHeight; y++){
				for(byte x = 0; x < nMaxWidth; x++){
					uint nTilePropSwatch = G.GetInt24(MapTileProp, 3 * ((y * nMaxWidth) + x));
					if((nTilePropSwatch & 0x002000) == 0x002000){
						graphMap.FillRectangle(UnkBrush, x << 4, y << 4, 16, 16);
					}
				}
			}
		}



		public void UpdateTilesInUse(byte nTile){
			ushort nCount = 0;
			for(byte z = 0; z < 3; z++) {
				if(LocForm.mnuLayer[z].Checked){
					for(byte y = 0; y < nMaxHeight; y++){
						for(byte x = 0; x < nMaxWidth; x++){
							if(MapData[nLOffset[z] + (y * nLWidth[z]) + x] == nTile) {
								nCount++;
							}
						}
					}
				}
			}
			// Worst case scenario, it takes about 20 times as long to update the whole screen as it does one tile
			if(nCount < 0x14) {
				for(byte z = 0; z < 3; z++) {
					if(LocForm.mnuLayer[z].Checked) {
						for(byte y = 0; y < nMaxHeight; y++) {
							for(byte x = 0; x < nMaxWidth; x++) {
								if(MapData[nLOffset[z] + (y * nLWidth[z]) + x] == nTile) {
									ChangeMapSection(y, (byte) (y + 1), x, (byte) (x + 1));
								}
							}
						}
					}
				}
			}
			else {
				RedrawMap();
			}
		}
	}
}
